@@ -10,7 +10,7 @@ import sys
 
 import setuptools
 
-from setupmeta import listify, MetaDefs, project_path, readlines, relative_path, Requirements, short, trace
+from setupmeta import get_words, listify, MetaDefs, project_path, readlines, relative_path, Requirements, short, trace, warn
 from setupmeta.content import find_contents, load_contents, load_list, load_readme, resolved_paths
 from setupmeta.license import determined_license
 from setupmeta.versioning import project_scm, Versioning
@@ -21,8 +21,6 @@ EXPLICIT = "explicit"
 CLASSIFIERS = "classifiers.txt"
 READMES = ["README.rst", "README.md", "README*"]
 
-RE_WORDS = re.compile(r"[^\w]+")
-
 # Accept reasonable variations of name + some separator + email
 RE_EMAIL = re.compile(r"(.+)[\s<>()\[\],:;]+([^@]+@[a-zA-Z0-9._-]+)")
 
@@ -321,7 +319,7 @@ def get_pip():
 
 def pythonified_name(name):
     if name:
-        words = [s.strip() for s in RE_WORDS.split(name)]
+        words = get_words(name)
         name = "_".join(s for s in words if s)
 
     return name
@@ -490,6 +488,12 @@ class SetupMeta(Settings):
                     SimpleModule("src", package, "__init__.py"),
                 )
 
+        if not self.name:
+            warn("'name' not specified in setup.py, auto-fill will be incomplete")
+
+        elif not self.definitions.get("packages") and not self.definitions.get("py_modules"):
+            warn("No 'packages' or 'py_modules' defined, this is an empty python package")
+
         scm = scm or project_scm(MetaDefs.project_dir)
         self.versioning = Versioning(self, scm)
         self.versioning.auto_fill_version()
